/* Font back-end driver for the NeXT/Open/GNUstep and MacOSX window system.
   See font.h

   Copyright (C) 2006 Free Software Foundation.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.

Author: Adrian Robert (arobert@cogsci.ucsd.edu)
*/

#include "config.h"
#ifdef USE_FONT_BACKEND

#include "lisp.h"
#include "dispextern.h"
#include "blockinput.h"
#include "charset.h"
#include "frame.h"
#include "window.h"
#include "fontset.h"
#include "nsterm.h"
#include "frame.h"
#include "character.h"
#include "font.h"

extern Lisp_Object prop_name_to_numeric(enum font_property_index prop,
                                        Lisp_Object name);

extern Lisp_Object Qns;
static Lisp_Object Qapple;
extern Lisp_Object ns_expand_space;
extern Lisp_Object Qappend;
extern int ns_antialias_text, ns_use_qd_smoothing;
extern float ns_antialias_threshold;
extern int ns_tmp_flags;

#define NSFONT_TRACE 0


// ============================================================================
//
//    Utilities
//
// ============================================================================


/* Replace spaces w/another character so emacs core font parsing routines
   aren't thrown off. */
static void
nsfont_escape_name(char *name)
{
  int i=0, len=strlen(name);
  for ( ; i<len; i++)
    if (name[i] == ' ')
      name[i] = '_';
}


/* Reconstruct spaces in a font family name passed through emacs. */
static void
nsfont_unescape_name(char *name)
{
  int i=0, len=strlen(name);
  for ( ; i<len; i++)
    if (name[i] == '_')
      name[i] = ' ';
}


/* Extract family name from a font spec. */
static NSString *
nsfont_get_family(Lisp_Object font_spec)
{
  Lisp_Object tem = AREF (font_spec, FONT_FAMILY_INDEX);
  if (NILP(tem))
      return nil;
  else
    {
      char *tmp = strdup(SDATA(SYMBOL_NAME(tem)));
      NSString *family;
      nsfont_unescape_name(tmp);
      //PENDING: this seems to be needed only for font names that are
      // hard-coded into emacs, like 'helvetica' for splash screen
      if (tmp)
        tmp[0] = toupper(tmp[0]);
      family = [NSString stringWithUTF8String: tmp];
      free(tmp);
      return family;
    }
}


/* Converts FONT_WEIGHT, FONT_SLANT, FONT_WIDTH to NSFont traits. */
static unsigned int
nsfont_spec_to_traits (Lisp_Object font_spec)
{
  Lisp_Object tem;
  unsigned int traits = 0;
  int n, ref;

  tem = AREF(font_spec, FONT_WEIGHT_INDEX);
  if (!NILP(tem))
    {
      n = XINT(tem);
      //PENDING: for now, distinguish medium and greater than medium
      ref = XINT(prop_name_to_numeric(FONT_WEIGHT_INDEX, intern("medium")));
//fprintf(stderr, "    got med wt ref = %d\n",ref);
      traits |= (n > ref) ? NSBoldFontMask : NSUnboldFontMask;
    }

  tem = AREF(font_spec, FONT_SLANT_INDEX);
  if (!NILP(tem))
    {
      n = XINT(tem);
      //PENDING: for now, distinguish roman and slanted
      ref = XINT(prop_name_to_numeric(FONT_SLANT_INDEX, intern("roman")));
//fprintf(stderr, "got roman slant ref = %d\n",ref);
      traits |= (n > ref) ? NSItalicFontMask : NSUnitalicFontMask;
    }

  tem = AREF(font_spec, FONT_WIDTH_INDEX);
  if (!NILP(tem))
    {
      n = XINT(tem);
      //PENDING: for now, distinguish normal, < normal, > normal
      ref = XINT(prop_name_to_numeric(FONT_WIDTH_INDEX, intern("normal")));
      if (n < ref)
        traits |= NSCondensedFontMask;
      else if (n > ref)
        traits |= NSExpandedFontMask;
    }

//fprintf(stderr, "  returning traits = %u\n",traits);
  return traits;
}


/* Converts NSArray of PS name, non-family part, weight, and traits to a
   font backend font-entity. */
static Lisp_Object
nsfont_fmember_to_entity (NSString *family, NSArray *famMember)
{
  Lisp_Object font_entity =
    Fmake_vector(make_number (FONT_ENTITY_MAX), null_string);
  unsigned int traits = [[famMember objectAtIndex: 3] unsignedIntValue];
//   NSString *psName = [famMember objectAtIndex: 0];
  NSMutableString *suffix = [[famMember objectAtIndex: 1] mutableCopy];
  char *escapedFamily = [family UTF8String];

  nsfont_escape_name(escapedFamily);
  [suffix replaceOccurrencesOfString: @" " withString: @"" options: nil
                               range: NSMakeRange(0, [suffix length])];

  ASET (font_entity, FONT_TYPE_INDEX, Qns);
  ASET (font_entity, FONT_FOUNDRY_INDEX, Qapple);
  ASET (font_entity, FONT_FAMILY_INDEX, intern(escapedFamily));
  ASET (font_entity, FONT_ADSTYLE_INDEX, intern([suffix UTF8String]));
  ASET (font_entity, FONT_REGISTRY_INDEX, Qiso10646_1);

  ASET (font_entity, FONT_WEIGHT_INDEX, traits & NSBoldFontMask ?
        prop_name_to_numeric(FONT_WEIGHT_INDEX, intern("bold")) :
        prop_name_to_numeric(FONT_WEIGHT_INDEX, intern("medium")));
  ASET (font_entity, FONT_SLANT_INDEX, traits & NSItalicFontMask ?
        prop_name_to_numeric(FONT_SLANT_INDEX, intern("italic")) :
        prop_name_to_numeric(FONT_SLANT_INDEX, intern("roman")));
  ASET (font_entity, FONT_WIDTH_INDEX, traits & NSCondensedFontMask ?
        prop_name_to_numeric(FONT_WIDTH_INDEX, intern("condensed")) :
        traits & NSExpandedFontMask ?
          prop_name_to_numeric(FONT_WIDTH_INDEX, intern("expanded")) :
          prop_name_to_numeric(FONT_WIDTH_INDEX, intern("normal")));

  ASET (font_entity, FONT_SIZE_INDEX, make_number(0));
  ASET (font_entity, FONT_EXTRA_INDEX, Qnil);
  ASET (font_entity, FONT_FRAME_INDEX, Qnil);
  ASET (font_entity, FONT_OBJLIST_INDEX, Qnil);

//   if (NSFONT_TRACE) {
//     fprintf(stderr, "created font_entity:\n    ");
//     debug_print(font_entity);
//   }
  return font_entity;
}


/* Computes Hamming distance btwn two "vectors" of 0's and 1's. */
static int
nsfont_trait_distance (unsigned int traits1, unsigned int traits2)
{
  int i, d=0;
  for (i=0; i<sizeof(unsigned int)*8; i++)
    {
      d += (traits1 & 0x1) ^ (traits2 & 0x1);
      traits1 >> 1;
      traits2 >> 1;
    }
  return d;
}


/* Default font entity based on Monaco. */
static Lisp_Object
nsfont_fallback_entity()
{
  NSString *family = [[NSFont userFixedPitchFontOfSize: 0] familyName];
  NSArray *famMemberSpec = [NSArray arrayWithObjects: family, @"",
                                    [NSNumber numberWithUnsignedInt: 5],
                                    [NSNumber numberWithUnsignedInt: 0], nil];
  return nsfont_fmember_to_entity(family, famMemberSpec);
}


// ============================================================================
//
//    Font driver implementation
//
// ============================================================================

static Lisp_Object nsfont_get_cache (Lisp_Object frame);
static Lisp_Object nsfont_list (Lisp_Object frame, Lisp_Object font_spec);
static Lisp_Object nsfont_match (Lisp_Object frame, Lisp_Object font_spec);
static Lisp_Object nsfont_list_family (Lisp_Object frame);
static struct font *nsfont_open (FRAME_PTR f, Lisp_Object font_entity,
                                 int pixel_size);
static void nsfont_close (FRAME_PTR f, struct font *font);
static int nsfont_has_char (Lisp_Object entity, int c);
static unsigned int nsfont_encode_char (struct font *font, int c);
static int nsfont_text_extents (struct font *font, unsigned int *code,
                                int nglyphs, struct font_metrics *metrics);
static int nsfont_draw (struct glyph_string *s, int from, int to, int x, int y,
                        int with_background);

struct font_driver nsfont_driver =
  {
    (Lisp_Object) NULL,		// Qns
    nsfont_get_cache,
    nsfont_list,
    nsfont_match,
    nsfont_list_family,
    NULL,			//free_entity
    nsfont_open,
    nsfont_close,
    NULL,			// prepare_face
    NULL,			// done_face
    nsfont_has_char,
    nsfont_encode_char,
    nsfont_text_extents,
    nsfont_draw,
    // excluded: get_bitmap, free_bitmap, get_outline, free_outline,
    //           anchor_point, otf_capability, otf_gsub, otf_gpos
  };


/* Return a cache of font-entities on FRAME.  The cache must be a
   cons whose cdr part is the actual cache area.  */
static Lisp_Object
nsfont_get_cache (Lisp_Object frame)
{
  Display_Info *dpyinfo = FRAME_NS_DISPLAY_INFO (XFRAME (frame));
  return (dpyinfo->name_list_element);
}


/* List fonts exactly matching with FONT_SPEC on FRAME.  The value
   is a vector of font-entities.  This is the sole API that
   allocates font-entities.  */
static Lisp_Object
nsfont_list (Lisp_Object frame, Lisp_Object font_spec)
{
  Lisp_Object list = Qnil;
  Lisp_Object tem;
  NSString *family;
  NSArray *families;
  NSEnumerator *famEnum;
  NSFontManager *fontMgr;
  unsigned int traits = nsfont_spec_to_traits(font_spec);

  if (NSFONT_TRACE)
    {
      fprintf(stderr, "nsfont: list for fontspec:\n    ");
      debug_print(font_spec);
    }

  // if has an adstyle or non-unicode registry, give up
  tem = AREF(font_spec, FONT_ADSTYLE_INDEX);
  if (!NILP(tem) && !EQ (tem, null_string))
    fprintf(stderr, "adstyle: '%s'\n",SDATA(tem));//return null_vector;
  tem = AREF(font_spec, FONT_REGISTRY_INDEX);
  if (!EQ(tem, Qiso10646_1) && !EQ(tem, Qunicode_bmp))
    return null_vector;

  fontMgr = [NSFontManager sharedFontManager];

  family = nsfont_get_family(font_spec);
  if (family != nil)
    families = [NSArray arrayWithObject: family];
  else
    families = [fontMgr availableFontFamilies];

  for (famEnum = [families objectEnumerator]; family = [famEnum nextObject]; )
    {
      NSEnumerator *fm;
      NSArray *fmember, *firstMember = nil;
      unsigned int mtraits;
      BOOL foundItal = NO;
      NSArray *famMembers = [fontMgr availableMembersOfFontFamily: family];
#ifdef COCOA
      // LastResort is special: not a family but a font name only
      if ([@"LastResort" isEqualToString: family] && [famMembers count] == 0)
        {
          famMembers = [NSArray arrayWithObject: [NSArray arrayWithObjects:
              @"LastResort", @"", [NSNumber numberWithUnsignedInt: 5],
              [NSNumber numberWithUnsignedInt: 0], nil]];
        }
#endif

      // fmember = [postscriptName style weight traits]
      fm  = [famMembers objectEnumerator];
      while (fmember = [fm nextObject])
        {
          mtraits = [[fmember objectAtIndex: 3] unsignedIntValue];
          if ((mtraits & traits) == traits)
            {
              list = Fcons (nsfont_fmember_to_entity(family, fmember), list);
              if (mtraits & NSItalicFontMask)
                foundItal = YES;
              if (firstMember == nil)
                firstMember = fmember;
            }
        }
      if (foundItal == NO && firstMember != nil)
        {
          // no italic member found; add a synthesized one
          NSMutableArray *smember = [firstMember mutableCopy];
          [smember replaceObjectAtIndex: 1 withObject: @"synthItal" ];
          mtraits = [[fmember objectAtIndex: 3] unsignedIntValue];
          mtraits |= NSItalicFontMask;
          [smember replaceObjectAtIndex: 3
                   withObject: [NSNumber numberWithUnsignedInt: mtraits]];
//NSLog(@"-- adding synthItal member: %@", smember);
          list = Fcons (nsfont_fmember_to_entity(family, smember), list);
        }
    }

  if (NSFONT_TRACE)
      fprintf(stderr, "    Returning %d entities.\n",XINT(Flength(list)));

  return (NILP (list) ? null_vector : Fvconcat (1, &list));
}


/* Return a font entity most closely maching with FONT_SPEC on
   FRAME.  The closeness is determined by the font backend, thus
   `face-font-selection-order' is ignored here.  */
static Lisp_Object
nsfont_match (Lisp_Object frame, Lisp_Object font_spec)
{
  long traits = nsfont_spec_to_traits(font_spec);
  NSFontManager *fontMgr = [NSFontManager sharedFontManager];
  NSString *family;
  Lisp_Object tem;

  if (NSFONT_TRACE)
    {
      fprintf(stderr, "nsfont: match for fontspec:\n    ");
      debug_print(font_spec);
    }

  // if has an adstyle or non-unicode registry, just return fallback
  tem = AREF(font_spec, FONT_ADSTYLE_INDEX);
  if (!NILP(tem) && !EQ (tem, null_string))
    fprintf(stderr, "adstyle: '%s'\n",SDATA(tem));//return nsfont_fallback_entity();
  tem = AREF(font_spec, FONT_REGISTRY_INDEX);
  if (!EQ(tem, Qiso10646_1) && !EQ(tem, Qunicode_bmp))
    return nsfont_fallback_entity();

  family = nsfont_get_family(font_spec);

  if (family != nil)
    {
      // try to find close font in family
      NSArray *famMembers = [fontMgr availableMembersOfFontFamily: family];
      NSEnumerator *fm = [famMembers objectEnumerator];
      NSArray *fmember;
      int minDist = sizeof(unsigned int) * 8 + 1;
      int bestMatchIdx = -1;
      int i;

      for (i=0; fmember = [fm nextObject]; i++)
        {
          unsigned int mtraits = [[fmember objectAtIndex: 3] unsignedIntValue];
          int dist = nsfont_trait_distance((mtraits & traits), traits);
          if (dist < minDist)
            {
              bestMatchIdx = i;
              minDist = dist;
            }
        }
      if (bestMatchIdx != -1)
        return nsfont_fmember_to_entity
          (family, [famMembers objectAtIndex: bestMatchIdx]);
    }

  // no family that had members was given; find any font matching traits
  {
    NSArray *fontNames = [fontMgr availableFontNamesWithTraits: traits];
    if (fontNames && [fontNames count]  > 0)
      {
        NSString *fontName = [fontNames objectAtIndex: 0];
        //PENDING: is there a more efficient way to get family name?
        NSFont *font = [NSFont fontWithName: fontName size: 0];
        if (font != nil)
          {
            // now need to get suffix part of name..
            NSString *family = [font familyName];
            NSEnumerator *fm = [[fontMgr availableMembersOfFontFamily: family]
                                 objectEnumerator];
            NSArray *fmember;
            while (fmember = [fm nextObject])
              {
                unsigned int mtraits =
                  [[fmember objectAtIndex: 3] unsignedIntValue];
                if (mtraits & traits == traits)
                  return nsfont_fmember_to_entity(family, fmember);
              }
          }
      }
  }

  // if we get here, return the fallback
  if (NSFONT_TRACE)
      fprintf(stderr, "    *** returning fallback\n");

  return nsfont_fallback_entity();
}


/* List available families.  The value is a list of family names
   (symbols). */
static Lisp_Object
nsfont_list_family (Lisp_Object frame)
{
  Lisp_Object list = Qnil;
  NSEnumerator *families =
    [[[NSFontManager sharedFontManager] availableFontFamilies]
      objectEnumerator];
  NSString *family;
  while (family = [families nextObject])
      list = Fcons(intern([family UTF8String]), list);
  //PENDING: escape the name?

  if (NSFONT_TRACE)
    fprintf(stderr, "nsfont: list families returning %d entries\n",
            XINT(Flength(list)));

  return list;
}


/* Open a font specified by FONT_ENTITY on frame F.  If the font is
   scalable, open it with PIXEL_SIZE.  */
static struct font *
nsfont_open (FRAME_PTR f, Lisp_Object font_entity, int pixel_size)
{
  BOOL synthItal;
  struct font *font;
  unsigned int traits = nsfont_spec_to_traits(font_entity);
  NSFontManager *fontMgr = [NSFontManager sharedFontManager];
  NSString *family;
  NSFont *nsfont, *sfont;
  Lisp_Object tem;

  if (NSFONT_TRACE)
    {
      fprintf(stderr, "nsfont: open size %d of fontentity:\n    ", pixel_size);
      debug_print(font_entity);
    }

  if (pixel_size <= 0)
    {
      // try to get it out of frame params
        Lisp_Object tem = get_frame_param(f, Qfontsize);
        pixel_size = NILP(tem) ? 0 : XFASTINT(tem);
    }

  tem = AREF (font_entity, FONT_ADSTYLE_INDEX);
  synthItal = !NILP(tem) && !strncmp("synthItal", SDATA(SYMBOL_NAME(tem)), 9);
  family = nsfont_get_family(font_entity);

  nsfont = [fontMgr fontWithFamily: family
                            traits: traits weight: 5 size: pixel_size];
  // if didn't find, try synthetic italic
  if (nsfont == nil && synthItal && (traits & NSItalicFontMask))
      nsfont = [fontMgr fontWithFamily: family
                                traits: traits & ~NSItalicFontMask
                                weight: 5 size: pixel_size];
#ifdef COCOA
  if ([@"LastResort" isEqualToString: family])
    {
      nsfont = [NSFont fontWithName: @"LastResort" size: pixel_size];
    }
#endif

  if (nsfont == nil)
    {
      message_with_string("*** Warning: font in family '%s' not found",
                          build_string([family UTF8String]), 1);
      nsfont = [NSFont userFixedPitchFontOfSize: pixel_size];
      if (!nsfont)
        {
          fprintf(stderr, "*** Emacs.app: unable to load backup font\n");
          return NULL;
        }
    }

  if (!(font = xmalloc(sizeof(struct font))))
    return NULL;

  // for metrics
  sfont = [nsfont screenFont];
  if (sfont == nil)
    sfont = nsfont;

  // backend font struct fields
  font->entity = font_entity;
  font->pixel_size = [sfont pointSize];
  font->driver = &nsfont_driver;
  font->scalable = 0;
  font->encoding_charset = -1;
  font->repertory_charset = -1;
  font->ascent = [sfont ascender];
  font->descent = -[sfont descender];
  font->file_name = NULL;
  font->min_width = [sfont widthOfString: @"|"]; // PENDING

  // "legacy" font structures:
  // "struct font" (font backend) contains "font_info" (emacs core) contains
  // "ns_font" (equiv to XFontStruct in other terms) contains NSFont
  {
    double expand, shrink, hshrink;
    float full_height, min_height, hd;
    const char *fontName = [[nsfont fontName] UTF8String];
    int len = strlen(fontName);
    struct font_info *fontInfo = &(font->font);
    struct ns_font *nsFont = (struct ns_font *)xmalloc(sizeof(struct ns_font));

    bzero (nsFont, sizeof(struct ns_font));

#ifdef GNUSTEP
    nsFont->nsfont = sfont;
#else
    nsFont->nsfont = nsfont;
#endif
    [nsFont->nsfont retain];

    // set up ns_font (defined in nsgui.h)
    nsFont->name = (char *)xmalloc(strlen(fontName)+1);
    bcopy(fontName, nsFont->name, strlen(fontName)+1);
    nsFont->ital =
      synthItal || ([fontMgr traitsOfFont:nsfont] & NSItalicFontMask);

    // Metrics etc.; some fonts return an unusually large max advance, so we
    // only use it for fonts that have wide characters.
    nsFont->width = [sfont numberOfGlyphs] > 1000 ?
      [sfont maximumAdvancement].width : [sfont widthOfString: @"0"];

    full_height = [sfont boundingRectForFont].size.height;
    min_height = [sfont ascender] - [sfont descender];
    hd = full_height - min_height;

    if (!NUMBERP(ns_expand_space))
      error("No expand space defined");

    // expand = 1.0 means use full_height, 0.0 use min_height, negative shrink
    expand = XFLOATINT(ns_expand_space);

    if (expand < 0.0)
      {
        shrink = 1 + expand;
        hshrink = 1 + expand / 2.0;
        expand = 0.0;
      }
    else
      shrink = hshrink = 1.0;

    nsFont->underpos = 2; //[sfont underlinePosition];//PENDING: often clipped out
    nsFont->underwidth = [sfont underlineThickness];
    nsFont->size = font->pixel_size;
    nsFont->voffset = rint(hshrink * [sfont ascender] + expand * hd / 2);

    // max bounds
    nsFont->max_bounds.ascent= rint(hshrink * [sfont ascender] + expand * hd/2);
    nsFont->max_bounds.descent= -rint(hshrink* [sfont descender] - expand*hd/2);
    nsFont->height = nsFont->max_bounds.ascent + nsFont->max_bounds.descent;
    nsFont->max_bounds.width = nsFont->width;
    nsFont->max_bounds.lbearing = 0;
    nsFont->max_bounds.rbearing =
      nsFont->width + (nsFont->ital ? 0.2 * nsFont->height : 0);

    // glyph and metrics caches: create a big string containing
    // all unicode characters
    {
      unichar *unichars = xmalloc(0x10001 * sizeof(unichar));
      int i;
      if (!unichars
          || !(nsFont->glyphs = xmalloc(0x10000 * sizeof(unsigned short)))
          || !(nsFont->widths = xmalloc(0x10000 * sizeof(unsigned char))))
        abort();

      unichars[0] = 0xFEFF;
      for (i=1; i<0x10000; i++)
        if (i < 0xD800 || i > 0xDFFF)
          unichars[i] = i;
        else
          unichars[i] = 0xFEFF;
      unichars[0x10000] = 0;

      {
#ifdef COCOA
        NSString *allChars = [[NSString alloc]
                               initWithCharactersNoCopy: unichars
                                                 length: 0x10001
                                           freeWhenDone: NO];
        EmacsGlyphStorage *glyphStorage =
          [[EmacsGlyphStorage alloc] initWithCapacity: 0x10001];
        NSGlyphGenerator *glyphGenerator =
          [NSGlyphGenerator sharedGlyphGenerator];
        NSCharacterSet *coveredChars = [nsfont coveredCharacterSet];
        unsigned int numGlyphs = [nsfont numberOfGlyphs];
        unsigned int gInd=0, cInd=0;

        [glyphStorage setString: allChars font: nsFont->nsfont];
        [glyphGenerator generateGlyphsForGlyphStorage: glyphStorage
                        desiredNumberOfCharacters: glyphStorage->maxChar
                                       glyphIndex: &gInd characterIndex: &cInd];
#endif
        for (i=0; i<0x10000; i++)
          {
            int g;
#ifdef COCOA
            g = glyphStorage->cglyphs[i];
            if (g > numGlyphs)
              g = 0xffff; // hopefully unused...
#else
            g = unichars[i+1];
#endif
            nsFont->glyphs[i] = g;
            nsFont->widths[g] = [sfont advancementForGlyph: g].width;
          }
#ifdef COCOA
        [glyphStorage release];
        [allChars release];
#endif
      }
      xfree(unichars);
    }

#ifdef COCOA
    // set up synthItal and the CG font
    nsFont->synthItal = synthItal;
    {
      ATSFontRef atsFont = ATSFontFindFromPostScriptName
        ((CFStringRef)[nsfont fontName], kATSOptionFlagsDefault);

      if (atsFont == kATSFontRefUnspecified)
        {
          // see if we can get it by dropping italic (then synthesizing)
          atsFont = ATSFontFindFromPostScriptName((CFStringRef)
              [[fontMgr convertFont:nsfont toNotHaveTrait:NSItalicFontMask]
                fontName], kATSOptionFlagsDefault);
          if (atsFont != kATSFontRefUnspecified)
              nsFont->synthItal = YES;
          else
            {
              // last resort fallback
              atsFont = ATSFontFindFromPostScriptName
                ((CFStringRef)@"Monaco", kATSOptionFlagsDefault);
            }
        }
      nsFont->cgfont = CGFontCreateWithPlatformFont((void*)&atsFont);
    }
#endif

    // set up font_info
    fontInfo->font = nsFont;
    nsFont->container = fontInfo;
    fontInfo->name = xmalloc(len + 1);
    if (! fontInfo->name)
        return NULL;
    bcopy(fontName, fontInfo->name, len + 1);
    fontInfo->full_name = xstrdup(fontInfo->name);
    fontInfo->size = nsFont->width;
    fontInfo->height = nsFont->height;
    fontInfo->charset = -1;
    fontInfo->space_width = [sfont widthOfString: @" "];
    fontInfo->average_width = nsFont->width;
    fontInfo->default_ascent = 0;
    fontInfo->vertical_centering = 0;
    fontInfo->baseline_offset = 0;
    fontInfo->relative_compose = 0;
    fontInfo->encoding_type = FONT_ENCODING_NOT_DECIDED;
    fontInfo->font_encoder = NULL;
  }
    return font;
}


/* Close FONT on frame F. */
static void
nsfont_close (FRAME_PTR f, struct font *font)
{
  struct font_info *fontInfo = &(font->font);
  struct ns_font *nsFont = fontInfo->font;
  xfree(nsFont->glyphs);
  xfree(nsFont->widths);
  [nsFont->nsfont release];
#ifdef COCOA
      CGFontRelease(nsFont->cgfont);
#endif
      xfree(nsFont->name);
      xfree(fontInfo->name);
      xfree(fontInfo->full_name);
      xfree(nsFont);
      free(font);
}


/* If FONT_ENTITY has a glyph for character C (Unicode code point),
   return 1.  If not, return 0.  If a font must be opened to check
   it, return -1. */
static int
nsfont_has_char (Lisp_Object entity, int c)
{
  return -1;
}


/* Return a glyph code of FONT for character C (Unicode code point).
   If FONT doesn't have such a glyph, return FONT_INVALID_CODE. */
static unsigned int
nsfont_encode_char (struct font *font, int c)
{
  struct font_info *fontInfo = &(font->font);
  unsigned short g;
  if (c > 0xFFFF)
    return FONT_INVALID_CODE;
  g = ((struct ns_font *)fontInfo->font)->glyphs[c];
//fprintf(stderr, "mapping char %d -> %d\n",c,g);
  return g == 0xFFFF ? FONT_INVALID_CODE : g;
}


/* Perform the size computation of glyphs of FONT and fill in members
   of METRICS.  The glyphs are specified by their glyph codes in
   CODE (length NGLYPHS). */
static int
nsfont_text_extents (struct font *font, unsigned int *code, int nglyphs,
                     struct font_metrics *metrics)
{
  struct ns_font *nsFont = font->font.font;
  XCharStruct max_bounds = nsFont->max_bounds;
  int i;
  float totalWidth = 0.0;

  for (i=0; i<nglyphs; i++)
    totalWidth += nsFont->widths[code[i]];

  metrics->lbearing = 0;
  metrics->ascent = max_bounds.ascent;
  metrics->descent = max_bounds.descent;
  metrics->width = totalWidth;
  metrics->rbearing = totalWidth + (nsFont->ital ? 0.2 * nsFont->height : 0);

  return totalWidth; // not specified in doc, but xfont.c does it
}


/* Draw glyphs between FROM and TO of S->char2b at (X Y) pixel
   position of frame F with S->FACE and S->GC.  If WITH_BACKGROUND
   is nonzero, fill the background in advance.  It is assured that
   WITH_BACKGROUND is zero when (FROM > 0 || TO < S->nchars). */
static int
nsfont_draw (struct glyph_string *s, int from, int to, int x, int y,
             int with_background)
// NOTE: focus and clip must be set
{
  static char cbuf[1024];
  char *c = cbuf;
#ifdef GNUSTEP
  static float advances[1024];
  float *adv = advances;
#else
  static CGSize advances[1024];
  CGSize *adv = advances;
#endif
  struct face *face = s->face;
  NSRect r;
  struct ns_font *font = (s->face->font == ~0 || s->face->font == 0)
    ? FRAME_FONT(s->f) : s->face->font;
  NSColor *col, *bgCol;
  unsigned short *t = s->char2b;
  int i, len;

  r.origin.x = s->x;
  if (s->row->full_width_p)
    r.origin.x += (WINDOW_HAS_VERTICAL_SCROLL_BAR_ON_LEFT(s->w) ? 1 : -1)
      * FRAME_INTERNAL_BORDER_WIDTH(s->f);
  if (s->face->box != FACE_NO_BOX && s->first_glyph->left_box_line_p)
    r.origin.x += abs(s->face->box_line_width);

  r.origin.y = s->y;
  r.size.height = s->height;

  // Convert UTF-16 (?) to UTF-8 and determine advances.  Note if we just ask
  // NS to render the string, it will come out differently from the individual
  // character widths added up because of layout processing.  Emacs would have
  // to pass off more text handling to the window system in order to work with
  // this, starting with a glyph_string_metrics() function..
  for (i=0; i<s->nchars; i++, t++)
    {
#ifdef GNUSTEP
      *adv++ = font->widths[*t];
      CHAR_STRING_ADVANCE(*t, c); // this converts the char to UTF-8
#else
      (*adv++).width = font->widths[*t];
#endif
    }
  len = adv - advances;
  r.size.width = s->background_width;
  *c = 0;

  if (s->row->full_width_p)
      r.size.width += 2 * FRAME_INTERNAL_BORDER_WIDTH(s->f);

  // fill background if requested
  if (with_background)
    {
      NSRect br = r;
      if (s->row->full_width_p && r.origin.y < 5)
        {
          // expand full-width top row over internal border
          br.origin.y -= FRAME_INTERNAL_BORDER_WIDTH(s->f);
          br.size.height += FRAME_INTERNAL_BORDER_WIDTH(s->f);
        }
      else if (s->row->mode_line_p || s->face->box != FACE_NO_BOX)
        {
          // don't overwrite box
          br.origin.y++;
          br.size.height -= 2;
        }
      if (s->face->box != FACE_NO_BOX)
        {
          if (s->first_glyph->left_box_line_p)
            br.origin.x -= abs(s->face->box_line_width);
          if ((s->first_glyph+ s->nchars-1)->right_box_line_p)
            br.size.width -= abs(s->face->box_line_width)/2;
        }
      if (!s->face->stipple)
        [(NS_FACE_BACKGROUND(face) != nil ?
          NS_FACE_BACKGROUND(face) : FRAME_BACKGROUND_COLOR(s->f)) set];
      else
        [[NSColor colorWithPatternImage:
                    [(EmacsImage *)face->stipple stippleRep]] set];
      NSRectFill(br);
    }

  // set up for character rendering
  r.origin.y += font->voffset + (r.size.height - font->height)/2;

  col = (NS_FACE_FOREGROUND(face) != nil ?
         NS_FACE_FOREGROUND(face) : FRAME_FOREGROUND_COLOR(s->f));
  //PENDING: find another way to pass this
  bgCol = (ns_tmp_flags != NS_DUMPGLYPH_FOREGROUND ? nil
           : (NS_FACE_BACKGROUND(face) != nil ?
              NS_FACE_BACKGROUND(face) : FRAME_BACKGROUND_COLOR(s->f)));

  // render under GNUstep using DPS
#ifdef GNUSTEP
  {
    NSGraphicsContext *context = GSCurrentContext();

    DPSgsave(context);
    [font->nsfont set];

    // do erase if "foreground" mode
    if (bgCol != nil)
      {
        [bgCol set];
        DPSmoveto(context, r.origin.x, r.origin.y);
//[context GSSetTextDrawingMode: GSTextFillStroke]; /// not implemented yet
        DPSxshow(context, cbuf, advances, len);
        DPSstroke(context);
        [col set];
//[context GSSetTextDrawingMode: GSTextFill]; /// not implemented yet
      }

    // do underline
    if (face->underline_p)
      {
        if (face->underline_color != nil)
          [(NSColor *)face->underline_color set];
        else
          [col set];
        DPSmoveto(context, r.origin.x, r.origin.y + font->underpos);
        DPSlineto(context, r.origin.x+r.size.width, r.origin.y+font->underpos);
        if (face->underline_color != nil)
          [col set];
      }
    else
      [col set];

    // draw with DPSxshow()
    DPSmoveto(context, r.origin.x, r.origin.y);
    DPSxshow(context, cbuf, advances, len);
    DPSstroke(context);

    DPSgrestore(context);
    return to-from;
  }

#else  // COCOA
  {
    CGContextRef gcontext =
      [[NSGraphicsContext currentContext] graphicsPort];
    static CGAffineTransform fliptf;
    static BOOL firstTime = YES;

    if (firstTime)
      {
        firstTime = NO;
        fliptf = CGAffineTransformMakeScale(1.0,-1.0);
      }

    CGContextSaveGState(gcontext);

    fliptf.c =  font->synthItal ? Fix2X(kATSItalicQDSkew) : 0.0;

    CGContextSetFont(gcontext, font->cgfont);
    CGContextSetFontSize(gcontext, font->size);
    if (ns_antialias_text == NO || font->size <= ns_antialias_threshold)
      CGContextSetShouldAntialias(gcontext, 0);
    else
      CGContextSetShouldAntialias(gcontext, 1);
    if (ns_use_qd_smoothing)
      CGContextSetFontRenderingMode(gcontext, 2); // 0 is Cocoa, 2 is QD

    CGContextSetTextMatrix(gcontext, fliptf);

    if (bgCol != nil)
      {
        // foreground drawing; erase first to avoid overstrike
        [bgCol set];
        CGContextSetTextDrawingMode(gcontext, kCGTextFillStroke);
        CGContextSetTextPosition(gcontext, r.origin.x, r.origin.y);
        CGContextShowGlyphsWithAdvances(gcontext, s->char2b, advances, len);
        CGContextSetTextDrawingMode(gcontext, kCGTextFill);
      }

    if (face->underline_p)
      {
        if (face->underline_color != nil)
          [(NSColor *)face->underline_color set];
        else
          [col set];
        CGContextBeginPath(gcontext);
        CGContextMoveToPoint(gcontext, r.origin.x, r.origin.y + font->underpos);
        CGContextAddLineToPoint(gcontext, r.origin.x + r.size.width,
                                r.origin.y + font->underpos);
        CGContextStrokePath(gcontext);
        if (face->underline_color != nil)
          [col set];
      }
    else
      [col set];

    CGContextSetTextPosition(gcontext, r.origin.x, r.origin.y);
    CGContextShowGlyphsWithAdvances(gcontext, s->char2b, advances, len);

    CGContextRestoreGState(gcontext);
    return;
  }
#endif  // COCOA

}


/*  Auto-creates a fontset built around the font in font_object,
    by creating an attributed string with characters from each
    script, then requesting the NS text system to fix attributes
    in range. */
void nsfont_make_fontset_for_font(Lisp_Object name, Lisp_Object font_object)
{
  Lisp_Object script, famAndReg;
  struct font *font_struct = XSAVE_VALUE(font_object)->pointer;
  struct ns_font *nsFont = font_struct->font.font;
  NSFont *font = nsFont->nsfont;

  // NS text system (and char buf) init
  static NSTextStorage *store;
  static NSLayoutManager *layout;
  static NSRange range;
  static NSMutableDictionary *attribs;
  static BOOL firstTime = YES;
  static Lisp_Object *scripts;
  Lisp_Object regString = build_string("iso10646-1");
  int i;

  if (firstTime == YES)
    {
      int nscripts = XINT(Flength(Vscript_representative_chars));
      unsigned char *buf = malloc(4*nscripts*sizeof(char));
      Lisp_Object scriptsChars = Vscript_representative_chars;
      unsigned char *tpos = buf;

      scripts = (Lisp_Object *) malloc(nscripts * sizeof(Lisp_Object));
      for (i=0; i<nscripts; i++)
        {
          Lisp_Object sChars = XCAR(scriptsChars);
          Lisp_Object chars = XCDR(sChars);
          unsigned int ch = XUINT(XCAR(chars));

          scripts[i] = XCAR(sChars);
          CHAR_STRING_ADVANCE(ch, tpos);
          scriptsChars = XCDR(scriptsChars);
        }
      *tpos = '\0';

      store = [[NSTextStorage alloc] init];
      layout = [[NSLayoutManager alloc] init];
      [store addLayoutManager: layout];
      [layout release];

      [store beginEditing];
      [[store mutableString] appendString:[NSString stringWithUTF8String: buf]];
      [store endEditing];

      free(buf);
      range = NSMakeRange(0, [store length]);

      if (range.length != nscripts)
        {
          //error
        }

      attribs = [[NSMutableDictionary alloc] init];
      firstTime = NO;
    }

  // set the fonts
  [store beginEditing];
  [store removeAttribute: NSFontAttributeName range: range];
  [attribs setObject: font forKey: NSFontAttributeName];
  [store addAttributes: attribs range: range];
  [store endEditing];
  [store ensureAttributesAreFixedInRange: range]; //PENDING: needed?

  // read them out
  for (i=0; i<range.length; i++)
    {
      NSFont *cfont = [store attribute: NSFontAttributeName atIndex: i
                        effectiveRange: NULL];
      char *family = [[cfont familyName] UTF8String];
      Lisp_Object famAndReg;

      nsfont_escape_name(family);
      famAndReg = Fcons(build_string(family), regString);

//       if (NSFONT_TRACE)
//         fprintf(stderr, "%s fontset: use '%s' for script '%s'\n",
//                 nsFont->name, family, SDATA(SYMBOL_NAME(scripts[i])));

      Fset_fontset_font (name, scripts[i], famAndReg, Qnil, Qnil);
    }

}


void
syms_of_nsfont ()
{
  nsfont_driver.type = Qns;
  register_font_driver (&nsfont_driver, NULL);
  DEFSYM (Qapple, "apple");
}

#endif /* USE_FONT_BACKEND */
